@@ -1,9 +1,10 @@
 from __future__ import annotations
 
+import contextlib
 import mmap
 import threading
+import warnings
 from enum import Enum
-from functools import lru_cache
 from itertools import product
 from pathlib import Path
 from typing import (
@@ -15,13 +16,14 @@ from typing import (
     SupportsInt,
     Union,
     cast,
+    no_type_check,
     overload,
 )
 
 import numpy as np
 
 from ._util import AXIS, VoxelSize, get_reader, is_supported_file
-from .structures import Attributes, ExpLoop, FrameMetadata, Metadata, XYPosLoop
+from .structures import ROI, Attributes, ExpLoop, FrameMetadata, Metadata, XYPosLoop
 
 try:
     from functools import cached_property
@@ -32,13 +34,19 @@ except ImportError:
 if TYPE_CHECKING:
     from typing import Any, Dict, List, Tuple
 
-    import dask.array as da
+    import dask.array.core
     import xarray as xr
     from typing_extensions import Literal
 
+    from ._sdk.latest import ND2Reader as LatestSDKReader
+    from .structures import Position
+
 
 Index = Union[int, slice]
 
+ROI_METADATA = "CustomData|RoiMetadata_v1"
+IMG_METADATA = "ImageMetadataLV"
+
 
 class ReadMode(str, Enum):
     MMAP = "mmap"
@@ -52,8 +60,10 @@ class ND2File:
     def __init__(
         self,
         path: Union[Path, str],
+        *,
         validate_frames: bool = False,
         search_window: int = 100,
+        read_using_sdk: bool = None,
     ) -> None:
         """Open an nd2 file.
 
@@ -61,17 +71,27 @@ class ND2File:
         ----------
         path : Union[Path, str]
             Filename of an nd2 file.
-        validate_frames : bool, optional
+        validate_frames : bool
             Whether to verify (and attempt to fix) frames whose positions have been
-            shifted relative to the predicted offset (i.e. in a corrupted file),
-            by default False.
-        search_window : int, optional
+            shifted relative to the predicted offset (i.e. in a corrupted file).
+            This comes at a slight performance penalty at file open, but may "rescue"
+            some corrupt files. by default False.
+        search_window : int
             When validate_frames is true, this is the search window (in KB) that will
             be used to try to find the actual chunk position. by default 100 KB
+        read_using_sdk : Optional[bool]
+            If `True`, use the SDK to read the file. If `False`, inspects the chunkmap
+            and reads from a `numpy.memmap`. If `None` (the default), uses the SDK if
+            the file is compressed, otherwise uses the memmap. Note: using
+            `read_using_sdk=False` on a compressed file will result in a ValueError.
+
         """
         self._path = str(path)
         self._rdr = get_reader(
-            self._path, validate_frames=validate_frames, search_window=search_window
+            self._path,
+            validate_frames=validate_frames,
+            search_window=search_window,
+            read_using_sdk=read_using_sdk,
         )
         self._closed = False
         self._is_legacy = "Legacy" in type(self._rdr).__name__
@@ -138,17 +158,132 @@ class ND2File:
         """Misc text info."""
         return self._rdr.text_info()
 
+    @cached_property
+    def rois(self) -> Dict[int, ROI]:
+        """Return dict of {id: ROI} for all ROIs found in the metadata."""
+        if self.is_legacy or ROI_METADATA not in self._rdr._meta_map:  # type: ignore
+            return {}
+        data = self.unstructured_metadata(include={ROI_METADATA})
+        data = data.get(ROI_METADATA, {}).get("RoiMetadata_v1", {})
+        data.pop("Global_Size", None)
+        try:
+            _rois = (ROI._from_meta_dict(d) for d in data.values())
+            rois = {r.id: r for r in _rois}
+        except Exception as e:
+            raise ValueError(f"Could not parse ROI metadata: {e}") from e
+        return rois
+
     @cached_property
     def experiment(self) -> List[ExpLoop]:
         """Loop information for each nd axis"""
-        return self._rdr.experiment()
+        exp = self._rdr.experiment()
+
+        # https://github.com/tlambert03/nd2/issues/78
+        # the SDK doesn't always do a good job of pulling position names from metadata
+        # here, we try to extract it manually.  Might be error prone, so currently
+        # we just ignore errors.
+        if not self.is_legacy and IMG_METADATA in self._rdr._meta_map:  # type: ignore
+            for n, item in enumerate(exp):
+                if isinstance(item, XYPosLoop):
+                    names = {
+                        tuple(p.stagePositionUm): p.name for p in item.parameters.points
+                    }
+                    if not any(names.values()):
+                        _exp = self.unstructured_metadata(
+                            include={IMG_METADATA}, unnest=True
+                        )[IMG_METADATA]
+                        if n >= len(_exp):
+                            continue
+                        with contextlib.suppress(Exception):
+                            _fix_names(_exp[n], item.parameters.points)
+        return exp
+
+    def unstructured_metadata(
+        self,
+        *,
+        unnest: bool = False,
+        strip_prefix: bool = True,
+        include: Optional[Set[str]] = None,
+        exclude: Optional[Set[str]] = None,
+    ) -> Dict[str, Any]:
+        """Exposes, and attempts to decode, each metadata chunk in the file.
+
+        This is provided as a *experimental* fallback in the event that
+        `ND2File.experiment` does not contain all of the information you need. No
+        attempt is made to parse or validate the metadata, and the format of various
+        sections, *may* change in future versions of nd2. Consumption of this metadata
+        should use appropriate exception handling!
+
+        The 'ImageMetadataLV' chunk is the most likely to contain useful information,
+        but if you're generally looking for "hidden" metadata, it may be helpful to
+        look at the full output.
+
+        Parameters
+        ----------
+        unnest : bool, optional
+            If `True` the nested `NextLevelEx` keys of each Experiment loop level will
+            be flattened into a list, and the return type will be `list`. by default
+            `False`.
+        strip_prefix : bool, optional
+            Whether to strip the type information from the front of the keys in the
+            dict. For example, if `True`: `uiModeFQ` becomes `ModeFQ` and `bUsePFS`
+            becomes `UsePFS`, etc... by default `True`
+        include : Optional[Set[str]], optional
+            If provided, only include the specified keys in the output. by default,
+            all metadata sections found in the file are included.
+        exclude : Optional[Set[str]], optional
+            If provided, exclude the specified keys from the output. by default `None`
+
+        Returns
+        -------
+        Dict[str, Any]
+            A dict of the unstructured metadata, with keys that are the type of the
+            metadata chunk (things like 'CustomData|RoiMetadata_v1' or
+            'ImageMetadataLV'), and values that are associated metadata chunk.
+        """
+        if self.is_legacy:
+            raise NotImplementedError(
+                "unstructured_metadata not available for legacy files"
+            )
+
+        from ._nd2decode import decode_metadata, unnest_experiments
+
+        output: Dict[str, Any] = {}
+
+        rdr = cast("LatestSDKReader", self._rdr)
+        keys = set(rdr._meta_map)
+        if include:
+            _keys: Set[str] = set()
+            for i in include:
+                if i not in keys:
+                    warnings.warn(f"Key {i!r} not found in metadata")
+                else:
+                    _keys.add(i)
+            keys = _keys
+        if exclude:
+            keys = {k for k in keys if k not in exclude}
+
+        for key in sorted(keys):
+            try:
+                meta: bytes = rdr._get_meta_chunk(key)
+                if meta.startswith(b"<"):
+                    # probably xml
+                    decoded: Any = meta.decode("utf-8")
+                else:
+                    decoded = decode_metadata(meta, strip_prefix=strip_prefix)
+                    if key == IMG_METADATA and unnest:
+                        decoded = unnest_experiments(decoded)
+            except Exception:
+                decoded = meta
+
+            output[key] = decoded
+        return output
 
     @cached_property
     def metadata(self) -> Union[Metadata, dict]:
         """Various metadata (will be dict if legacy format)."""
         return self._rdr.metadata()
 
-    @lru_cache(maxsize=1024)
     def frame_metadata(
         self, seq_index: Union[int, tuple]
     ) -> Union[FrameMetadata, dict]:
@@ -156,7 +291,18 @@ class ND2File:
 
         This includes the global metadata from the metadata function.
         (will be dict if legacy format).
+
+        Parameters
+        ----------
+        seq_index : Union[int, tuple]
+            frame index
+
+        Returns
+        -------
+        Union[FrameMetadata, dict]
+            dict if legacy format, else FrameMetadata
         """
+
         idx = cast(
             int,
             self._seq_index_from_coords(seq_index)
@@ -228,7 +374,18 @@ class ND2File:
         return np.dtype(f"{d}{attrs.bitsPerComponentInMemory // 8}")
 
     def voxel_size(self, channel: int = 0) -> VoxelSize:
-        """XYZ voxel size."""
+        """XYZ voxel size.
+
+        Parameters
+        ----------
+        channel : int
+            Channel for which to retrieve voxel info, by default 0
+
+        Returns
+        -------
+        VoxelSize
+            Named tuple with attrs `x`, `y`, and `z`.
+        """
         return VoxelSize(*self._rdr.voxel_size())
 
     def asarray(self, position: Optional[int] = None) -> np.ndarray:
@@ -245,8 +402,10 @@ class ND2File:
 
         Raises
         ------
+        ValueError
+            if `position` is a string and is not a valid position name
         IndexError
-            if position is provided and is out of range
+            if `position` is provided and is out of range
         """
         final_shape = list(self.shape)
         if position is None:
@@ -290,7 +449,7 @@ class ND2File:
         """array protocol"""
         return self.asarray()
 
-    def to_dask(self, wrapper=True, copy=True) -> da.Array:
+    def to_dask(self, wrapper=True, copy=True) -> dask.array.core.Array:
         """Create dask array (delayed reader) representing image.
 
         This generally works well, but it remains to be seen whether performance
@@ -305,13 +464,13 @@ class ND2File:
         wrapper : bool
             If True (the default), the returned obect will be a thin subclass of
             a :class:`dask.array.Array` (an
-            `ResourceBackedDaskArray`) that manages the opening
-            and closing of this file when getting chunks via compute(). If `wrapper`
-            is `False`, then a pure `da.Array` will be returned. However, when that
-            array is computed, it will incur a file open/close on *every* chunk
-            that is read (in the `_dask_block` method).  As such `wrapper`
-            will generally be much faster, however, it *may* fail (i.e. result in
-            segmentation faults) with certain dask schedulers.
+            `ResourceBackedDaskArray`) that manages the opening and closing of this file
+            when getting chunks via compute(). If `wrapper` is `False`, then a pure
+            `dask.array.core.Array` will be returned. However, when that array is
+            computed, it will incur a file open/close on *every* chunk that is read (in
+            the `_dask_block` method).  As such `wrapper` will generally be much faster,
+            however, it *may* fail (i.e. result in segmentation faults) with certain
+            dask schedulers.
         copy : bool
             If `True` (the default), the dask chunk-reading function will return
             an array copy. This can avoid segfaults in certain cases, though it
@@ -319,7 +478,7 @@ class ND2File:
 
         Returns
         -------
-        da.Array
+        dask.array.core.Array
         """
         from dask.array import map_blocks
 
@@ -365,8 +524,9 @@ class ND2File:
                             f"Cannot get chunk {block_id} for single frame image."
                         )
                     idx = 0
-                data = self._get_frame(cast(int, idx))[(np.newaxis,) * ncoords]
-                return data.copy() if copy else data
+                data = self._get_frame(int(idx))  # type: ignore
+                data = data.copy() if copy else data
+                return data[(np.newaxis,) * ncoords]
             finally:
                 if was_closed:
                     self.close()
@@ -380,6 +540,10 @@ class ND2File:
     ) -> xr.DataArray:
         """Create labeled xarray representing image.
 
+        `array.dims` will be populated according to image metadata, and coordinates
+        will be populated based on pixel spacings. Additional metadata is available
+        in `array.attrs['metadata']`.
+
         Parameters
         ----------
         delayed : bool
@@ -459,24 +623,35 @@ class ND2File:
     def _frame_count(self) -> int:
         return int(np.prod(self._coord_shape))
 
-    def _get_frame(self, index: int) -> np.ndarray:
-        frame = self._rdr._read_image(index)
+    def _get_frame(self, index: SupportsInt) -> np.ndarray:
+        frame = self._rdr._read_image(int(index))
         frame.shape = self._raw_frame_shape
         return frame.transpose((2, 0, 1, 3)).squeeze()
 
-    def _expand_coords(self, squeeze=True) -> dict:
-        """Return a dict that can be used as the coords argument to xr.DataArray"""
+    def _expand_coords(self, squeeze: bool = True) -> dict:
+        """Return a dict that can be used as the coords argument to xr.DataArray
+
+        Parameters
+        ----------
+        squeeze : bool
+            whether to squeeze axes with length < 2, by default True
+
+        Returns
+        -------
+        dict
+            dict of axis name -> coordinates
+        """
         dx, dy, dz = self.voxel_size()
 
         coords: Dict[str, Sized] = {
             AXIS.Y: np.arange(self.attributes.heightPx) * dy,
-            AXIS.X: np.arange(self.attributes.widthPx) * dx,
+            AXIS.X: np.arange(self.attributes.widthPx or 1) * dx,
             AXIS.CHANNEL: self._channel_names,
             AXIS.POSITION: ["XYPos:0"],  # maybe overwritten below
         }
 
         for c in self.experiment:
-            if squeeze and getattr(c, "count") <= 1:
+            if squeeze and c.count <= 1:
                 continue
             if c.type == "ZStackLoop":
                 coords[AXIS.Z] = np.arange(c.count) * c.parameters.stepUm
@@ -526,26 +701,104 @@ class ND2File:
 
 @overload
 def imread(
-    file: str, dask: Literal[False] = False, xarray: Literal[False] = False
+    file: Union[Path, str],
+    *,
+    dask: Literal[False],
+    xarray: Literal[False],
+    validate_frames: bool = False,
+    read_using_sdk: Optional[bool] = None,
 ) -> np.ndarray:
     ...
 
 
 @overload
-def imread(file: str, dask: bool = ..., xarray: Literal[True] = True) -> xr.DataArray:
+def imread(
+    file: Union[Path, str],
+    *,
+    dask: bool = ...,
+    xarray: Literal[True],
+    validate_frames: bool = False,
+    read_using_sdk: Optional[bool] = None,
+) -> xr.DataArray:
     ...
 
 
 @overload
-def imread(file: str, dask: Literal[True] = ..., xarray=False) -> da.Array:
+def imread(
+    file: Union[Path, str],
+    *,
+    dask: Literal[True],
+    xarray: Literal[False],
+    validate_frames: bool = False,
+    read_using_sdk: Optional[bool] = None,
+) -> dask.array.core.Array:
     ...
 
 
-def imread(file: str, dask: bool = False, xarray: bool = False):
-    with ND2File(file) as nd2:
+def imread(
+    file: Union[Path, str],
+    *,
+    dask: bool = False,
+    xarray: bool = False,
+    validate_frames: bool = False,
+    read_using_sdk: Optional[bool] = None,
+):
+    """Open `file`, return requested array type, and close `file`.
+
+    Parameters
+    ----------
+    file : Union[Path, str]
+        Filepath (`str`) or `Path` object to ND2 file.
+    dask : bool
+        If `True`, returns a (delayed) `dask.array.Array`. This will avoid reading
+        any data from disk until specifically requested by using `.compute()` or
+        casting to a numpy array with `np.asarray()`. By default `False`.
+    xarray : bool
+        If `True`, returns an `xarray.DataArray`, `array.dims` will be populated
+        according to image metadata, and coordinates will be populated based on pixel
+        spacings. Additional metadata is available in `array.attrs['metadata']`.
+        If `dask` is also `True`, will return an xarray backed by a delayed dask array.
+        By default `False`.
+    validate_frames : bool
+        Whether to verify (and attempt to fix) frames whose positions have been
+        shifted relative to the predicted offset (i.e. in a corrupted file).
+        This comes at a slight performance penalty at file open, but may "rescue"
+        some corrupt files. by default False.
+    read_using_sdk : Optional[bool]
+        If `True`, use the SDK to read the file. If `False`, inspects the chunkmap and
+        reads from a `numpy.memmap`. If `None` (the default), uses the SDK if the file
+        is compressed, otherwise uses the memmap.
+        Note: using `read_using_sdk=False` on a compressed file will result in a
+        ValueError.
+
+    Returns
+    -------
+    Union[np.ndarray, dask.array.Array, xarray.DataArray]
+        Array subclass, depending on arguments used.
+    """
+    with ND2File(
+        file, validate_frames=validate_frames, read_using_sdk=read_using_sdk
+    ) as nd2:
         if xarray:
             return nd2.to_xarray(delayed=dask)
         elif dask:
             return nd2.to_dask()
         else:
             return nd2.asarray()
+
+
+@no_type_check
+def _fix_names(xy_exp, points: List[Position]) -> None:
+    """Attempt to fix missing XYPosLoop position names."""
+    if not isinstance(xy_exp, dict) or xy_exp.get("Type", "") != 2:
+        raise ValueError("Invalid XY experiment")
+    _points = xy_exp["LoopPars"]["Points"]
+    if len(_points) == 1 and "" in _points:
+        _points = _points[""]
+    if not isinstance(_points, list):
+        _points = [_points]
+    _names = {(p["PosX"], p["PosY"], p["PosZ"]): p["PosName"] for p in _points}
+
+    for p in points:
+        if p.name is None:
+            p.name = _names.get(tuple(p.stagePositionUm), p.name)
